@@ -18,47 +18,16 @@ TableHelper.reserved = reserved_words
 TableHelper.operators = OPERATORS
 
 
-def _get_table_helper(tx, table):
-    """
-    Utility function to create a TableHelper instance.
-    Ensures consistent configuration across all SQL methods.
-    """
-    return TableHelper(tx, table)
-
-
-def _validate_table_name(table):
-    """Validate table name format."""
-    if not table or not isinstance(table, str):
-        raise ValueError("Table name must be a non-empty string")
-    # Add more validation as needed
-    return table.strip()
-
-
-def _handle_predicate_errors(predicates, operation="WHERE"):
-    """Process a list of predicates with error handling."""
-    sql_parts = []
-    vals = []
-
-    for pred, val in predicates:
-        sql_parts.append(pred)
-        if val is None:
-            pass
-        elif isinstance(val, tuple):
-            vals.extend(val)
-        else:
-            vals.append(val)
-
-    return sql_parts, vals
-
-
 system_fields = [
     "sys_id",
     "sys_created",
     "sys_modified",
     "sys_modified_by",
+    "sys_modified_row",
+    "sys_modified_count",
     "sys_dirty",
     "sys_table",
-    "description",
+    "sys_keywords",
 ]
 
 
@@ -143,7 +112,7 @@ class SQL(BaseSQLDialect):
         vals = []
 
         # Create table helper instance
-        th = _get_table_helper(tx, table)
+        th = TableHelper(tx, table)
 
         # Handle columns and DISTINCT before aliasing
         if columns is None:
@@ -447,7 +416,7 @@ class SQL(BaseSQLDialect):
         if not isinstance(data, Mapping) or not data:
             raise ValueError("data must be a non-empty mapping of column-value pairs.")
 
-        th = _get_table_helper(tx, table)
+        th = TableHelper(tx, table)
         set_clauses = []
         vals = []
 
@@ -595,51 +564,164 @@ class SQL(BaseSQLDialect):
 
     @classmethod
     def merge(cls, tx, table, data, pk, on_conflict_do_nothing, on_conflict_update):
+        if not isinstance(data, Mapping) or not data:
+            raise ValueError("data must be a non-empty mapping of column-value pairs.")
+
+        table_helper = TableHelper(tx, table)
+        data = dict(data)  # work with a copy to avoid mutating the caller's dict
+
         if pk is None:
             pkeys = tx.table(table).primary_keys()
             if not pkeys:
                 raise ValueError("Primary key required for merge.")
-            # If there are multiple primary keys, use all of them
-            if len(pkeys) > 1:
-                pk = {pk: data[pk] for pk in pkeys}
-            else:
-                pk = {pkeys[0]: data[pkeys[0]]}
-            # Remove primary keys from data; they will be used in the conflict target
-            data = {k: v for k, v in data.items() if k not in pk}
-
-        # Create a merged dictionary for insert (data + primary key columns)
-        full_data = {}
-        full_data.update(data)
-        full_data.update(pk)
-
-        sql, vals = cls.insert(table, full_data)
-        sql = [sql]
-        vals = list(vals)  # Convert to a mutable list
-
-        if on_conflict_do_nothing != on_conflict_update:
-            sql.append("ON CONFLICT")
-            sql.append("(")
-            sql.append(",".join(pk.keys()))
-            sql.append(")")
-            sql.append("DO")
-            if on_conflict_do_nothing:
-                sql.append("NOTHING")
-            elif on_conflict_update:
-                # Call update() with excluded=True to produce the SET clause for the upsert.
-                sql_update, vals_update = cls.update(tx, table, data, pk, excluded=True)
-                sql.append(sql_update)
-                # Use list.extend to add the update values to vals.
-                vals.extend(vals_update)
+            missing = [key for key in pkeys if key not in data]
+            if missing:
+                missing_cols = ", ".join(missing)
+                raise ValueError(
+                    "Primary key values missing from data for merge: "
+                    f"{missing_cols}. Provide pk=... or include the key values in data."
+                )
+            pk = {key: data[key] for key in pkeys}
         else:
+            pk = dict(pk)
+            for key, value in pk.items():
+                if key in data and data[key] != value:
+                    raise ValueError(
+                        f"Conflicting values for primary key '{key}' between data and pk arguments."
+                    )
+
+        insert_data = dict(data)
+        insert_data.update(pk)
+
+        update_data = {k: v for k, v in data.items() if k not in pk}
+
+        if not update_data and on_conflict_update:
+            # Nothing to update, fall back to a no-op on conflict resolution.
+            on_conflict_do_nothing = True
+            on_conflict_update = False
+
+        if on_conflict_do_nothing == on_conflict_update:
             raise Exception(
                 "Update on conflict must have one and only one option to complete on conflict."
             )
 
+        sql, vals = cls.insert(table, insert_data)
+        sql = [sql]
+        vals = list(vals)  # Convert to a mutable list
+
+        sql.append("ON CONFLICT")
+        conflict_columns = [TableHelper.quote(column) for column in pk.keys()]
+        sql.append("(")
+        sql.append(", ".join(conflict_columns))
+        sql.append(")")
+        sql.append("DO")
+        if on_conflict_do_nothing:
+            sql.append("NOTHING")
+        elif on_conflict_update:
+            sql_update, vals_update = cls.update(
+                tx, table, update_data, pk, excluded=True
+            )
+            sql.append(sql_update)
+            vals.extend(vals_update)
+
         import sqlparse
 
         final_sql = sqlparse.format(" ".join(sql), reindent=True, keyword_case="upper")
         return final_sql, tuple(vals)
 
+    @classmethod
+    def insnx(cls, tx, table, data, where=None):
+        """Insert only when the supplied predicate finds no existing row."""
+        if not table:
+            raise ValueError("Table name is required.")
+        if not isinstance(data, Mapping) or not data:
+            raise ValueError("data must be a non-empty mapping of column-value pairs.")
+
+        # Helper used for quoting and foreign key resolution
+        th = TableHelper(tx, table)
+        quote_helper = TableHelper(None, table)
+
+        columns_sql = []
+        select_parts = []
+        vals = []
+
+        for key, val in data.items():
+            columns_sql.append(quote_helper.quote(key.lower()))
+            if isinstance(val, str) and len(val) > 2 and val.startswith("@@") and val[2:]:
+                select_parts.append(val[2:])
+            else:
+                select_parts.append("%s")
+                vals.append(val)
+
+        if not select_parts:
+            raise ValueError("At least one column is required for insert.")
+
+        if where is None:
+            if tx is None:
+                raise ValueError(
+                    "A transaction context is required when deriving WHERE from primary keys."
+                )
+            pk_cols = tx.table(table).primary_keys()
+            if not pk_cols:
+                raise ValueError("Primary key required to derive WHERE clause.")
+            missing = [pk for pk in pk_cols if pk not in data]
+            if missing:
+                raise ValueError(
+                    "Missing primary key value(s) for insert condition: " + ", ".join(missing)
+                )
+            where = {pk: data[pk] for pk in pk_cols}
+
+        where_clauses = []
+        where_vals = []
+
+        if isinstance(where, Mapping):
+            compiled = []
+            for key, val in where.items():
+                compiled.append(th.make_predicate(key, val))
+            where = compiled
+
+        if isinstance(where, str):
+            where_clauses.append(where)
+        else:
+            try:
+                for predicate, value in where:
+                    where_clauses.append(predicate)
+                    if value is None:
+                        continue
+                    if isinstance(value, tuple):
+                        where_vals.extend(value)
+                    else:
+                        where_vals.append(value)
+            except (TypeError, ValueError) as exc:
+                raise ValueError(
+                    "Invalid WHERE clause format. Expected mapping, SQL string, or iterable of predicate/value pairs."
+                ) from exc
+
+        vals.extend(where_vals)
+
+        exists_sql = [
+            "SELECT 1 FROM",
+            TableHelper.quote(table),
+        ]
+        if where_clauses:
+            exists_sql.append("WHERE " + " AND ".join(where_clauses))
+
+        sql_parts = [
+            "INSERT INTO",
+            TableHelper.quote(table),
+            f"({','.join(columns_sql)})",
+            "SELECT",
+            ", ".join(select_parts),
+            "WHERE NOT EXISTS (",
+            " ".join(exists_sql),
+            ")",
+        ]
+
+        final_sql = sqlparse.format(" ".join(sql_parts), reindent=True, keyword_case="upper")
+        return final_sql, tuple(vals)
+
+    insert_if_not_exists = insnx
+
     @classmethod
     def version(cls):
         return "select version()", tuple()
@@ -719,14 +801,56 @@ class SQL(BaseSQLDialect):
     def drop_database(cls, name):
         return f"drop database if exists {name}", tuple()
 
+    @staticmethod
+    def _sys_modified_function_sql(schema_identifier):
+        return f"""
+            CREATE OR REPLACE FUNCTION {schema_identifier}.on_sys_modified()
+              RETURNS TRIGGER AS
+            $BODY$
+            BEGIN
+                IF (TG_OP = 'INSERT') THEN
+                    NEW.sys_table := TG_TABLE_NAME;
+                    NEW.sys_created := transaction_timestamp();
+                    NEW.sys_modified := transaction_timestamp();
+                    NEW.sys_modified_row := clock_timestamp();
+                    NEW.sys_modified_count := 0;
+                ELSIF (TG_OP = 'UPDATE') THEN
+                    NEW.sys_table := TG_TABLE_NAME;
+                    NEW.sys_created := OLD.sys_created;
+                    NEW.sys_modified_count := COALESCE(OLD.sys_modified_count, 0);
+                    IF ROW(NEW.*) IS DISTINCT FROM ROW(OLD.*) THEN
+                        IF OLD.sys_dirty IS TRUE AND NEW.sys_dirty IS FALSE THEN
+                            NEW.sys_dirty := FALSE;
+                        ELSE
+                            NEW.sys_dirty := TRUE;
+                        END IF;
+                        NEW.sys_modified := transaction_timestamp();
+                        NEW.sys_modified_row := clock_timestamp();
+                        NEW.sys_modified_count := COALESCE(OLD.sys_modified_count, 0) + 1;
+                    END IF;
+                END IF;
+                RETURN NEW;
+            END;
+            $BODY$
+              LANGUAGE plpgsql VOLATILE
+              COST 100;
+        """
+
     @classmethod
     def create_table(cls, name, columns={}, drop=False):
         if "." in name:
             fqtn = TableHelper.quote(name)
         else:
             fqtn = f"public.{TableHelper.quote(name)}"
+
         schema, table = fqtn.split(".")
-        name = fqtn.replace(".", "_")
+        schema_unquoted = schema.replace('"', "")
+        table_unquoted = table.replace('"', "")
+        trigger_name = (
+            f"on_update_row_{schema_unquoted}_{table_unquoted}".replace(".", "_")
+        )
+        trigger_identifier = TableHelper.quote(trigger_name)
+        schema_identifier = TableHelper.quote(schema_unquoted)
         sql = []
         if drop:
             sql.append(cls.drop_table(fqtn)[0])
@@ -734,39 +858,25 @@ class SQL(BaseSQLDialect):
             f"""
             CREATE TABLE {fqtn} (
               sys_id BIGSERIAL PRIMARY KEY,
-              sys_modified TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
               sys_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
-              sys_modified_by TEXT,
+              sys_modified TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
+              sys_modified_by TEXT NOT NULL DEFAULT 'SYSTEM',
+              sys_modified_row TIMESTAMP NOT NULL DEFAULT CLOCK_TIMESTAMP(),
+              sys_modified_count INTEGER NOT NULL DEFAULT 0,
               sys_dirty BOOLEAN NOT NULL DEFAULT FALSE,
-              sys_table TEXT,
-              description TEXT
+              sys_table TEXT NOT NULL,
+                            sys_keywords TEXT
             );
 
             SELECT SETVAL(PG_GET_SERIAL_SEQUENCE('{fqtn}', 'sys_id'),1000,TRUE);
 
-            CREATE OR REPLACE FUNCTION {schema}.on_sys_modified()
-              RETURNS TRIGGER AS
-            $BODY$
-                        BEGIN
-                        -- update sys_modified on each insert/update.
-                        NEW.sys_modified := now();
-                        if (TG_OP = 'INSERT') THEN
-                            NEW.sys_created :=now();
-                        ELSEIF (TG_OP = 'UDPATE') THEN
-                         -- Do not allow sys_created to be modified.
-                            NEW.sys_created := OLD.sys_created;
-                        END IF;
-                        -- Insert table name to row
-                        NEW.sys_table := TG_TABLE_NAME;
-                        RETURN NEW;
-                        END;
-            $BODY$
-              LANGUAGE plpgsql VOLATILE
-              COST 100;
+            {cls._sys_modified_function_sql(schema_identifier)}
 
-            CREATE TRIGGER on_update_row_{fqtn.replace('.', '_')}
+            DROP TRIGGER IF EXISTS {trigger_identifier} ON {fqtn};
+
+            CREATE TRIGGER {trigger_identifier}
             BEFORE INSERT OR UPDATE ON {fqtn}
-            FOR EACH ROW EXECUTE PROCEDURE {schema}.on_sys_modified();
+            FOR EACH ROW EXECUTE PROCEDURE {schema_identifier}.on_sys_modified();
 
         """
         )
@@ -782,6 +892,178 @@ class SQL(BaseSQLDialect):
         sql = sqlparse.format(" ".join(sql), reindent=True, keyword_case="upper")
         return sql, tuple()
 
+    @classmethod
+    def ensure_system_columns(
+        cls,
+        name,
+        existing_columns=None,
+        force=False,
+        existing_indexes=None,
+    ):
+        """Ensure all Velocity system columns and triggers exist for the table."""
+        existing_columns = {
+            col.lower() for col in (existing_columns or [])
+        }
+
+        existing_indexes = {
+            (idx or "").lower(): (definition or "")
+            for idx, definition in (existing_indexes or {}).items()
+        }
+
+        required_columns = [
+            "sys_id",
+            "sys_created",
+            "sys_modified",
+            "sys_modified_by",
+            "sys_modified_row",
+            "sys_modified_count",
+            "sys_dirty",
+            "sys_table",
+            "sys_keywords",
+        ]
+
+        missing_columns = [
+            column for column in required_columns if column not in existing_columns
+        ]
+
+        if not missing_columns and not force:
+            return None
+
+        if "." in name:
+            schema_name, table_name = name.split(".", 1)
+        else:
+            schema_name = cls.default_schema
+            table_name = name
+
+        schema_unquoted = schema_name.replace('"', "")
+        table_unquoted = table_name.replace('"', "")
+
+        schema_identifier = TableHelper.quote(schema_unquoted)
+        table_identifier = TableHelper.quote(table_unquoted)
+        fqtn = f"{schema_identifier}.{table_identifier}"
+
+        trigger_name = (
+            f"on_update_row_{schema_unquoted}_{table_unquoted}".replace(".", "_")
+        )
+        trigger_identifier = TableHelper.quote(trigger_name)
+
+        statements = []
+        added_columns = set()
+        columns_after = set(existing_columns)
+
+        if "sys_id" in missing_columns:
+            statements.append(
+                f"ALTER TABLE {fqtn} ADD COLUMN {TableHelper.quote('sys_id')} BIGSERIAL PRIMARY KEY;"
+            )
+            added_columns.add("sys_id")
+            columns_after.add("sys_id")
+
+        column_definitions = {
+            "sys_created": "TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP",
+            "sys_modified": "TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP",
+            "sys_modified_by": "TEXT NOT NULL DEFAULT 'SYSTEM'",
+            "sys_modified_row": "TIMESTAMP NOT NULL DEFAULT CLOCK_TIMESTAMP()",
+            "sys_modified_count": "INTEGER NOT NULL DEFAULT 0",
+            "sys_dirty": "BOOLEAN NOT NULL DEFAULT FALSE",
+            "sys_table": "TEXT",
+            "sys_keywords": "TEXT",
+        }
+
+        for column, definition in column_definitions.items():
+            if column in missing_columns:
+                statements.append(
+                    f"ALTER TABLE {fqtn} ADD COLUMN {TableHelper.quote(column)} {definition};"
+                )
+                added_columns.add(column)
+                columns_after.add(column)
+
+        default_map = {
+            "sys_created": "CURRENT_TIMESTAMP",
+            "sys_modified": "CURRENT_TIMESTAMP",
+            "sys_modified_by": "'SYSTEM'",
+            "sys_modified_row": "CLOCK_TIMESTAMP()",
+            "sys_modified_count": "0",
+            "sys_dirty": "FALSE",
+        }
+
+        for column, default_sql in default_map.items():
+            if column in columns_after and (force or column in added_columns):
+                quoted_column = TableHelper.quote(column)
+                statements.append(
+                    f"UPDATE {fqtn} SET {quoted_column} = {default_sql} WHERE {quoted_column} IS NULL;"
+                )
+                statements.append(
+                    f"ALTER TABLE {fqtn} ALTER COLUMN {quoted_column} SET DEFAULT {default_sql};"
+                )
+
+        if "sys_table" in columns_after and (force or "sys_table" in added_columns):
+            quoted_column = TableHelper.quote("sys_table")
+            table_literal = table_unquoted.replace("'", "''")
+            statements.append(
+                f"UPDATE {fqtn} SET {quoted_column} = COALESCE({quoted_column}, '{table_literal}') WHERE {quoted_column} IS NULL;"
+            )
+
+        not_null_columns = {
+            "sys_created",
+            "sys_modified",
+            "sys_modified_by",
+            "sys_modified_row",
+            "sys_modified_count",
+            "sys_dirty",
+            "sys_table",
+        }
+
+        for column in not_null_columns:
+            if column in columns_after and (force or column in added_columns):
+                statements.append(
+                    f"ALTER TABLE {fqtn} ALTER COLUMN {TableHelper.quote(column)} SET NOT NULL;"
+                )
+
+        reset_trigger = force or bool(added_columns)
+
+        if reset_trigger:
+            statements.append(
+                f"DROP TRIGGER IF EXISTS {trigger_identifier} ON {fqtn};"
+            )
+            statements.append(cls._sys_modified_function_sql(schema_identifier))
+            statements.append(
+                f"""
+                CREATE TRIGGER {trigger_identifier}
+                BEFORE INSERT OR UPDATE ON {fqtn}
+                FOR EACH ROW EXECUTE PROCEDURE {schema_identifier}.on_sys_modified();
+            """
+            )
+
+        if "sys_keywords" in columns_after:
+            base_index_name = f"idx_{schema_unquoted}_{table_unquoted}_sys_keywords_tsv"
+            base_index_name = re.sub(r"[^0-9a-zA-Z_]+", "_", base_index_name)
+            if len(base_index_name) > 60:
+                digest = hashlib.sha256(base_index_name.encode("utf-8")).hexdigest()
+                base_index_name = f"idx_{table_unquoted[:30]}_{digest[:8]}_sk"
+                base_index_name = re.sub(r"[^0-9a-zA-Z_]+", "_", base_index_name)
+            index_name = base_index_name[:63]
+            quoted_index_name = TableHelper.quote(index_name)
+
+            target_expression = "to_tsvector('simple', coalesce(sys_keywords, ''))"
+            current_definition = existing_indexes.get(index_name.lower(), "").lower()
+            needs_rebuild = force or target_expression not in current_definition
+
+            if needs_rebuild:
+                statements.append(
+                    f"DROP INDEX IF EXISTS {quoted_index_name};"
+                )
+                statements.append(
+                    f"CREATE INDEX {quoted_index_name} ON {fqtn} USING GIN (TO_TSVECTOR('simple', COALESCE(sys_keywords, '')));"
+                )
+
+        if not statements:
+            return None
+
+        sql = sqlparse.format(
+            " ".join(statements), reindent=True, keyword_case="upper"
+        )
+        return sql, tuple()
+
     @classmethod
     def drop_table(cls, name):
         return f"drop table if exists {TableHelper.quote(name)} cascade;", tuple()
@@ -1259,6 +1541,19 @@ class SQL(BaseSQLDialect):
         """
         Returns SQL for retrieving all indexes on a given table with detailed attributes.
         """
+        if "." in table:
+            schema, tbl = table.split(".", 1)
+            schema = schema.replace('"', "")
+            tbl = tbl.replace('"', "")
+            return (
+                """
+                SELECT indexname, tablename, schemaname, indexdef
+                FROM pg_indexes
+                WHERE tablename = %s AND schemaname = %s
+                """,
+                (tbl, schema),
+            )
+
         return (
             """
             SELECT indexname, tablename, schemaname, indexdef
